import LimitedArray from "./LimitedArray.js";
import Task from "./Task.js";
import Log from './Log.js'
import LogFmt from "./LogFmt.js";

export default class Cache {

  static create(key) {
    return new Cache(key)
  }

  constructor(cacheKey) {
    this.key = cacheKey;
    this.storage = new Map();
    this.keysArray = new LimitedArray(999_999_999);

    this._seconds = null;
    this._limit = null;

    this.totalRequests = 0;
    this.totalHits = 0;
    this.totalMisses = 0;
  }

  get(key) {
    const value = this.storage.get(key);

    this.totalRequests++;
    if (value) {
      this.totalHits++;
      Log.debug(`Using '${this.key}' cache`);
    } else {
      this.totalMisses++;
    }

    return value;
  }

  add(key, value) {
    const existingValue = this.get(key);
    if (!existingValue) {
      this.keysArray.add(key);
      Log.debug(`Adding to '${this.key}' cache`);
      this._deleteIfLastRemoved();
    }
    this.storage.set(key, value);
  }

  delete(key) {
    this.storage.delete(key);
  }

  limit(limit) {
    this._limit = limit;
    this.keysArray.limit = limit;
    return this;
  }

  hours(hours) {
    return this.minutes(hours * 60);
  }

  minutes(minutes) {
    return this.seconds(minutes * 60)
  }

  seconds(seconds) {
    return this._clearSeconds(seconds);
  }

  milliseconds(milliseconds) {
    const seconds = Math.round(milliseconds / 1000);
    return this._clearSeconds(seconds);
  }

  logInit() {
    const logFmt = LogFmt.msg('Init cache')
      .tag('cache').string('cache-name', this.key);
    if (this._limit) {
      logFmt.number('cache-limit', this._limit);
    }
    if (this._seconds) {
      logFmt.number('cache-seconds', this._seconds);
    }

    Log.info(logFmt.log());
  }

  _deleteIfLastRemoved() {
    const lastRemovedKey = this.keysArray.getLastRemoved();
    if (this._limit && lastRemovedKey) {
      this.delete(lastRemovedKey);
    }
  }

  _clearCache() {
    this.storage.clear();
    this.keysArray.clear();
  }

  _clearSeconds(seconds) {
    this._seconds = seconds;
    Task.do(() => {
      this._clearCache();

      Log.trace(
        LogFmt.msg('Clearing cache')
          .tag('cache')
          .string('state', 'clear')
          .string('cache_name', this.key)
          .log()
      );

    }).every(seconds);
    return this;
  }

}
