import axios from 'axios';
import Log from './Log.js';
import LogFmt from './LogFmt.js';

export default class Request {

  url;
  method;
  data;
  headers;

  response;
  statusCode;
  responseData;
  responseHeaders;

  startMs;

  constructor(url, method, data = null, headers = null) {
    this.url = url;
    this.method = method;
    this.data = data;
    this.headers = headers;
  }

  // This can be override.
  _addNeccesaryHeaders() {
    return this.headers;
  }

  // This can be override.
  _shouldIgnore() {
    return false;
  }

  async _onError(error) {
    // This can be override.
  }

  async _request() {
    this.startMs = Date.now();

    try {
      const config = this._generateRequestConfig();

      this.response = await axios(config);
      this.responseData = this.response.data;
      this.statusCode = this.response.status;
      this.responseHeaders = this.response.headers;

      Log.debug(
        LogFmt.msg('Request finished')
          .tag('request')
          .string('method', this.method)
          .number('status', this.statusCode)
          .string('url', this.url)
          .decimal('elapsed_ms', Date.now() - this.startMs)
          .log()
      );
      // FIXME: Pretty print. Log.debug(`Headers: ${JSON.stringify(this.responseHeaders)}`);
    } catch (error) {
      this.response = this._findError(error);
      this.responseHeaders = this.response.headers;
      this.statusCode = this.response.status;

      if (!this._shouldIgnore()) {
        await this._onError(error);
        this._processError();
      }
    }
    return this;
  }

  _generateRequestConfig() {
    let config = {
      method: this.method,
      url: this.url,
      headers: this._addNeccesaryHeaders()
    };
    if (this.data) {
      config.data = JSON.stringify(this.data);
    }
    return config;
  }

  _findError(error) {
    if (error.response) {
      return error.response;
    } else if (error.request) {
      return error.request;
    }
    return error;
  }

  _processError() {
    const logFmt = LogFmt.msg('Request failed').tag('request')
      .string('method', this.method)
      .string('url', this.url);

    if (this.response.status) {
      logFmt
        .number('status', this.statusCode)
      // FIXME: Pretty print.
      // Log.error(`Data: ${JSON.stringify(this.response.data)}`);
      // Log.error(`Headers: ${JSON.stringify(this.response.headers)}`);
    } else {
      logFmt
        .string('unknown_error', this._processStringError());
    }
    Log.error(logFmt
      .decimal('elapsed_ms', Date.now() - this.startMs).log());
  }

  _processStringError() {
    let stringError;
    try {
      stringError = JSON.stringify(this.response);
    } catch (error) {
      stringError = this.response;
    }
    return stringError;
  }

}
